package de.unikoblenz.fgbks.base.utils.boundary.impl;

import static de.unikoblenz.fgbks.base.utils.boundary.bicreater.BoundaryBiCreaterType.APPEND;
import static de.unikoblenz.fgbks.base.utils.boundary.bicreater.BoundaryBiCreaterType.BETWEEN;
import static de.unikoblenz.fgbks.base.utils.boundary.bicreater.BoundaryBiCreaterType.COMBINE;
import static de.unikoblenz.fgbks.base.utils.boundary.bicreater.BoundaryBiCreaterType.INTERSECTION;
import static de.unikoblenz.fgbks.base.utils.boundary.bicreater.BoundaryBiCreaterType.LOWER_BOUNDS;
import static de.unikoblenz.fgbks.base.utils.boundary.bicreater.BoundaryBiCreaterType.UPPER_BOUNDS;
import static de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheckType.IS_EQUAL;
import static de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheckType.IS_IN_CONTACT;
import static de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheckType.IS_NOT_IN_CONTACT;
import static de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheckType.IS_OVERLAPPING;
import static de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheckType.IS_SUBSUMPTION;
import static de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheckType.SUBSUMES;
import static de.unikoblenz.fgbks.base.utils.boundary.creater.BoundaryCreaterType.LOWER;
import static de.unikoblenz.fgbks.base.utils.boundary.creater.BoundaryCreaterType.UPPER;

import de.unikoblenz.fgbks.base.builder.DefaultBuilder;
import de.unikoblenz.fgbks.base.utils.boundary.AbstractBoundary;
import de.unikoblenz.fgbks.base.utils.boundary.Boundary;
import de.unikoblenz.fgbks.base.utils.boundary.BoundaryParseException;
import de.unikoblenz.fgbks.base.utils.boundary.bicreater.BiCreaterStringAppend;
import de.unikoblenz.fgbks.base.utils.boundary.bicreater.BiCreaterStringBetween;
import de.unikoblenz.fgbks.base.utils.boundary.bicreater.BiCreaterStringCombine;
import de.unikoblenz.fgbks.base.utils.boundary.bicreater.BiCreaterStringIntersection;
import de.unikoblenz.fgbks.base.utils.boundary.bicreater.BiCreaterStringLowerBounds;
import de.unikoblenz.fgbks.base.utils.boundary.bicreater.BiCreaterStringUpperBounds;
import de.unikoblenz.fgbks.base.utils.boundary.bicreater.BoundaryBiCreater;
import de.unikoblenz.fgbks.base.utils.boundary.bicreater.BoundaryBiCreaterType;
import de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheck;
import de.unikoblenz.fgbks.base.utils.boundary.checker.BoundaryCheckType;
import de.unikoblenz.fgbks.base.utils.boundary.checker.CheckStringEqual;
import de.unikoblenz.fgbks.base.utils.boundary.checker.CheckStringInContact;
import de.unikoblenz.fgbks.base.utils.boundary.checker.CheckStringNotInContact;
import de.unikoblenz.fgbks.base.utils.boundary.checker.CheckStringOverlap;
import de.unikoblenz.fgbks.base.utils.boundary.checker.CheckStringSubsumes;
import de.unikoblenz.fgbks.base.utils.boundary.checker.CheckStringSubsumption;
import de.unikoblenz.fgbks.base.utils.boundary.creater.BoundaryCreater;
import de.unikoblenz.fgbks.base.utils.boundary.creater.BoundaryCreaterType;
import de.unikoblenz.fgbks.base.utils.boundary.creater.CreaterStringLower;
import de.unikoblenz.fgbks.base.utils.boundary.creater.CreaterStringUpper;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.stream.Collectors;
import org.apache.commons.lang3.Validate;

public class StringBoundary extends AbstractBoundary<String> {

  private static HashMap<BoundaryCheckType, BoundaryCheck> checkerMap = new HashMap<>();
  private static HashMap<BoundaryCreaterType, BoundaryCreater> createrMap = new HashMap<>();
  private static HashMap<BoundaryBiCreaterType, BoundaryBiCreater> biCreaterMap = new HashMap<>();

  static {
    checkerMap.put(IS_EQUAL, CheckStringEqual.getInstance());
    checkerMap.put(IS_IN_CONTACT, CheckStringInContact.getInstance());
    checkerMap.put(IS_NOT_IN_CONTACT, CheckStringNotInContact.getInstance());
    checkerMap.put(IS_OVERLAPPING, CheckStringOverlap.getInstance());
    checkerMap.put(IS_SUBSUMPTION, CheckStringSubsumption.getInstance());
    checkerMap.put(SUBSUMES, CheckStringSubsumes.getInstance());
    createrMap.put(UPPER, CreaterStringUpper.getInstance());
    createrMap.put(LOWER, CreaterStringLower.getInstance());
    biCreaterMap.put(INTERSECTION, BiCreaterStringIntersection.getInstance());
    biCreaterMap.put(LOWER_BOUNDS, BiCreaterStringLowerBounds.getInstance());
    biCreaterMap.put(UPPER_BOUNDS, BiCreaterStringUpperBounds.getInstance());
    biCreaterMap.put(BETWEEN, BiCreaterStringBetween.getInstance());
    biCreaterMap.put(APPEND, BiCreaterStringAppend.getInstance());
    biCreaterMap.put(COMBINE, BiCreaterStringCombine.getInstance());
  }

  private boolean matchNoneOfValues;
  private boolean matchAny;
  private String[] values;
  // sorted list of cached hashes of valuesHashes
  private int[] valuesHashes;

  public StringBoundary(String text) {
    super(text, String.class);
  }

  private StringBoundary() {
    super(String.class);
  }

  @Override
  public int compareTo(Boundary o) {
    return o.getText().compareTo(o.getText());
  }

  public boolean matchesNoneOfValues() {
    return matchNoneOfValues;
  }

  public boolean isMatchAny() {
    return matchAny;
  }

  public String[] getValues() {
    return values.clone(); // performance ?
  }

  public int[] getValuesHashes() {
    return valuesHashes.clone(); // performance ?
  }

  public int getAmountOfElements() {
    return values.length;
  }

  @Override
  protected void parse(String text) {
    matchNoneOfValues = false;
    matchAny = false;
    try {
      matchAny = Validate.notNull(text).isEmpty();
      if (matchAny) {
        values = new String[0];
      } else {
        matchNoneOfValues = text.toLowerCase().startsWith("not");
        text = text.replace('(', '\0').replace(')', '\0').trim();
        if (matchNoneOfValues) {
          text = text.replaceFirst("not", "").trim();
        }
        values = text.split(",");
      }
      calcHashes();
    } catch (Exception e) {
      throw new BoundaryParseException("String value is not valid: " + this.getText());
    }
  }

  private void calcHashes() {
    valuesHashes = new int[values.length];
    for (int i = 0; i < values.length; i++) {
      values[i] = values[i].replace('"', '\0').trim();
      valuesHashes[i] = Objects.hashCode(values[i]);
    }
    // Sort hashes for later comparisons
    Arrays.sort(valuesHashes);
  }

  @Override
  protected void validate() {
    Validate.noNullElements(values);
  }

  @Override
  protected Map<BoundaryCheckType, BoundaryCheck> getChecker() {
    return Collections.unmodifiableMap(checkerMap);
  }

  @Override
  protected Map<BoundaryBiCreaterType, BoundaryBiCreater> getBiCreater() {
    return Collections.unmodifiableMap(biCreaterMap);
  }

  @Override
  protected Map<BoundaryCreaterType, BoundaryCreater> getCreater() {
    return Collections.unmodifiableMap(createrMap);
  }

  @Override
  public String getParsedText() {
    return (matchNoneOfValues ? "not(" : "")
        + Arrays.stream(values).map(v -> '"' + v + '"').collect(Collectors.joining(","))
        + (matchNoneOfValues ? ")" : "");
  }

  public static Builder getBuilder() {
    return new StringBoundary().new Builder();
  }

  public class Builder extends DefaultBuilder<StringBoundary> {

    private Set<String> buildingValues;

    private Builder() {
      buildingValues = new HashSet<>();
      value.matchAny = false;
      value.matchNoneOfValues = false;
    }

    public Builder isMatchNoneOfValues(boolean matchNoneOfValues) {
      value.matchNoneOfValues = matchNoneOfValues;
      return this;
    }

    public Builder isMatchAny(boolean matchAny) {
      value.matchAny = matchAny;
      return this;
    }

    public Builder addValue(String value) {
      buildingValues.add(Validate.notNull(value));
      return this;
    }

    public Builder addValues(Collection<String> values) {
      buildingValues.addAll(Validate.noNullElements(values));
      return this;
    }

    public Builder addValues(String[] values) {
      return addValues(Validate.notNull(Arrays.asList(values)));
    }

    @Override
    public StringBoundary build() {
      value.values = buildingValues.toArray(new String[0]);
      if (value.matchAny) {
        value.matchNoneOfValues = false;
        value.values = new String[0];
      }
      value.calcHashes();
      value.text = value.getParsedText();
      return super.build();
    }
  }
}
